from svvamp import Rule, RulePlurality, Profile, RuleVeto, RuleMaximin, RuleExhaustiveBallot, RuleBaldwin, \
    RuleIRVAverage, RuleMajorityJudgment, RuleNanson, RuleRankedPairs, RuleBorda, RuleSchulze, RuleIRVDuels, \
    RuleRangeVoting, RuleKemeny, RuleICRV, RuleCondorcetSumDefeats, OPTIONS


def test_initialize_options():
    """
        >>> rule = Rule(unexpected_option=42)
        Traceback (most recent call last):
        ValueError: ('Unknown option:', 'unexpected_option')
    """
    pass


def test_mean_utility_w():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [1, 0, 2],
        ...     [1, 2, 0],
        ...     [2, 1, 0],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RulePlurality()(profile)
        >>> rule.mean_utility_w_
        1.4
    """
    pass


def test_demo_profile():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [1, 0, 2],
        ...     [1, 2, 0],
        ...     [2, 1, 0],
        ...     [2, 1, 0],
        ... ])
        >>> rule = Rule()(profile)
        >>> rule.demo_profile_()  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        *********************
        *                   *
        *   Profile Class   *
        *                   *
        *********************
        <BLANKLINE>
        ************************
        *   Basic properties   *
        ************************
        n_v = 5
        n_c = 3
        labels_candidates = ['0', '1', '2']
        preferences_ut =
        [[2 1 0]
         [1 2 0]
         [0 2 1]
         [0 1 2]
         [0 1 2]]
        preferences_borda_ut =
        [[2. 1. 0.]
         [1. 2. 0.]
         [0. 2. 1.]
         [0. 1. 2.]
         [0. 1. 2.]]
        preferences_borda_rk =
        [[2 1 0]
         [1 2 0]
         [0 2 1]
         [0 1 2]
         [0 1 2]]
        preferences_rk =
        [[0 1 2]
         [1 0 2]
         [1 2 0]
         [2 1 0]
         [2 1 0]]
        PROFILE: Compute v_has_same_ordinal_preferences_as_previous_voter
        v_has_same_ordinal_preferences_as_previous_voter =
        [False False False False  True]
        <BLANKLINE>
        ************************
        *   Plurality scores   *
        ************************
        preferences_rk (reminder) =
        [[0 1 2]
         [1 0 2]
         [1 2 0]
         [2 1 0]
         [2 1 0]]
        PROFILE: Compute Plurality scores (rk)
        plurality_scores_rk = [1 2 2]
        PROFILE: Compute majority favorite (rk)
        majority_favorite_rk = nan
        PROFILE: Compute majority favorite (rk_ctb)
        majority_favorite_rk_ctb = nan
        <BLANKLINE>
        preferences_borda_ut (reminder) =
        [[2. 1. 0.]
         [1. 2. 0.]
         [0. 2. 1.]
         [0. 1. 2.]
         [0. 1. 2.]]
        PROFILE: Compute Plurality scores (ut)
        plurality_scores_ut = [1 2 2]
        PROFILE: Compute majority favorite (ut)
        majority_favorite_ut = nan
        PROFILE: Compute majority favorite (ut_ctb)
        majority_favorite_ut_ctb = nan
        <BLANKLINE>
        ********************
        *   Borda scores   *
        ********************
        preferences_borda_rk (reminder) =
        [[2 1 0]
         [1 2 0]
         [0 2 1]
         [0 1 2]
         [0 1 2]]
        PROFILE: Compute Borda scores of the candidates (rankings)
        PROFILE: Compute matrix of duels (with strict orders)
        borda_score_c_rk =
        [3 7 5]
        Remark: Borda scores above are computed with the matrix of duels.
        Check: np.sum(self.preferences_borda_rk, 0) =
        [3 7 5]
        PROFILE: Compute decreasing_borda_scores_rk
        decreasing_borda_scores_rk =
        [7 5 3]
        PROFILE: Compute candidates_by_decreasing_borda_score_rk
        candidates_by_decreasing_borda_score_rk =
        [1 2 0]
        <BLANKLINE>
        preferences_borda_ut (reminder) =
        [[2. 1. 0.]
         [1. 2. 0.]
         [0. 2. 1.]
         [0. 1. 2.]
         [0. 1. 2.]]
        PROFILE: Compute Borda scores of the candidates (weak orders)
        borda_score_c_ut =
        [3. 7. 5.]
        PROFILE: Compute decreasing_borda_scores_ut
        decreasing_borda_scores_ut =
        [7. 5. 3.]
        PROFILE: Compute candidates_by_decreasing_borda_score_ut
        candidates_by_decreasing_borda_score_ut =
        [1 2 0]
        <BLANKLINE>
        *****************
        *   Utilities   *
        *****************
        preferences_ut (reminder) =
        [[2 1 0]
         [1 2 0]
         [0 2 1]
         [0 1 2]
         [0 1 2]]
        PROFILE: Compute total utility of candidates
        total_utility_c =
        [3 7 5]
        PROFILE: Compute total_utility_min
        total_utility_min = 3
        PROFILE: Compute total_utility_max
        total_utility_max = 7
        PROFILE: Compute total_utility_mean
        total_utility_mean = 5.0
        PROFILE: Compute total_utility_std
        total_utility_std = 1.632993161855452
        <BLANKLINE>
        *******************************************
        *   Condorcet notions based on rankings   *
        *******************************************
        preferences_rk (reminder) =
        [[0 1 2]
         [1 0 2]
         [1 2 0]
         [2 1 0]
         [2 1 0]]
        matrix_duels_rk =
        [[0 1 2]
         [4 0 3]
         [3 2 0]]
        PROFILE: Compute matrix_victories_rk
        matrix_victories_rk =
        [[0. 0. 0.]
         [1. 0. 1.]
         [1. 0. 0.]]
        PROFILE: Compute condorcet_winner_rk
        condorcet_winner_rk = 1
        PROFILE: Compute exists_condorcet_order_rk
        exists_condorcet_order_rk = True
        PROFILE: Compute matrix_victories_rk_ctb
        matrix_victories_rk_ctb =
        [[0. 0. 0.]
         [1. 0. 1.]
         [1. 0. 0.]]
        PROFILE: Compute condorcet_winner_rk_ctb
        condorcet_winner_rk_ctb = 1
        PROFILE: Compute exists_condorcet_order_rk_ctb
        exists_condorcet_order_rk_ctb = True
        <BLANKLINE>
        ***************************************
        *   Relative Condorcet notions (ut)   *
        ***************************************
        preferences_borda_ut (reminder) =
        [[2. 1. 0.]
         [1. 2. 0.]
         [0. 2. 1.]
         [0. 1. 2.]
         [0. 1. 2.]]
        PROFILE: Compute matrix of duels
        matrix_duels_ut =
        [[0 1 2]
         [4 0 3]
         [3 2 0]]
        PROFILE: Compute matrix_victories_ut_rel
        matrix_victories_ut_rel =
        [[0. 0. 0.]
         [1. 0. 1.]
         [1. 0. 0.]]
        PROFILE: Compute condorcet_winner_ut_rel
        condorcet_winner_ut_rel = 1
        PROFILE: Compute exists_condorcet_order_ut_rel
        exists_condorcet_order_ut_rel = True
        PROFILE: Compute matrix_victories_ut_rel_ctb
        matrix_victories_ut_rel_ctb =
        [[0. 0. 0.]
         [1. 0. 1.]
         [1. 0. 0.]]
        PROFILE: Compute condorcet_winner_ut_rel_ctb
        condorcet_winner_ut_rel_ctb = 1
        PROFILE: Compute exists_condorcet_order_ut_rel_ctb
        exists_condorcet_order_ut_rel_ctb = True
        <BLANKLINE>
        ***************************************
        *   Absolute Condorcet notions (ut)   *
        ***************************************
        matrix_duels_ut (reminder) =
        [[0 1 2]
         [4 0 3]
         [3 2 0]]
        PROFILE: Compute matrix_victories_ut_abs
        matrix_victories_ut_abs =
        [[0. 0. 0.]
         [1. 0. 1.]
         [1. 0. 0.]]
        PROFILE: Compute Condorcet-admissible candidates
        condorcet_admissible_candidates =
        [False  True False]
        PROFILE: Compute number of Condorcet-admissible candidates
        nb_condorcet_admissible = 1
        PROFILE: Compute weak Condorcet winners
        weak_condorcet_winners =
        [False  True False]
        PROFILE: Compute number of weak Condorcet winners
        nb_weak_condorcet_winners = 1
        PROFILE: Compute Condorcet winner
        condorcet_winner_ut_abs = 1
        PROFILE: Compute exists_condorcet_order_ut_abs
        exists_condorcet_order_ut_abs = True
        PROFILE: Compute Resistant Condorcet winner
        resistant_condorcet_winner = nan
        PROFILE: Compute threshold_c_prevents_w_Condorcet_ut_abs
        threshold_c_prevents_w_Condorcet_ut_abs =
        [[0 0 1]
         [1 0 2]
         [0 1 0]]
        PROFILE: Compute matrix_victories_ut_abs_ctb
        matrix_victories_ut_abs_ctb =
        [[0. 0. 0.]
         [1. 0. 1.]
         [1. 0. 0.]]
        PROFILE: Compute condorcet_winner_ut_abs_ctb
        condorcet_winner_ut_abs_ctb = 1
        PROFILE: Compute exists_condorcet_order_ut_abs_ctb
        exists_condorcet_order_ut_abs_ctb = True
        <BLANKLINE>
        **********************************************
        *   Implications between Condorcet notions   *
        **********************************************
        maj_fav_ut (False)             ==>            maj_fav_ut_ctb (False)
         ||          ||                                     ||           ||
         ||          V                                      V            ||
         ||         maj_fav_rk (False) ==> maj_fav_rk_ctb (False)        ||
         V                         ||       ||                           ||
        Resistant Condorcet (False)                                      ||
         ||                        ||       ||                           ||
         V                         ||       ||                           V
        Condorcet_ut_abs (True)        ==>      Condorcet_ut_abs_ctb (True)
         ||          ||            ||       ||              ||           ||
         ||          V             V        V               V            ||
         ||       Condorcet_rk (True)  ==> Condorcet_rk_ctb (True)       ||
         V                                                               V
        Condorcet_ut_rel (True)        ==>      Condorcet_ut_rel_ctb (True)
         ||
         V
        Weak Condorcet (True)
         ||
         V
        Condorcet-admissible (True)
    """
    pass


def test_iia_subset_maximum_size_setter():
    """
        >>> rule = Rule()
        >>> rule.iia_subset_maximum_size = 'a'
        Traceback (most recent call last):
        ValueError: Unknown value for iia_subset_maximum_size: a (number or np.inf expected).
    """
    pass


def test_im_option_setter():
    """
        >>> rule = Rule()
        >>> rule.im_option = 'a'
        Traceback (most recent call last):
        ValueError: Unknown value for im_option: a
    """
    pass


def test_tm_option_setter():
    """
        >>> rule = Rule()
        >>> rule.tm_option = 'a'
        Traceback (most recent call last):
        ValueError: Unknown value for tm_option: a
    """
    pass


def test_um_option_setter():
    """
        >>> rule = Rule()
        >>> rule.um_option = 'a'
        Traceback (most recent call last):
        ValueError: Unknown value for um_option: a
    """
    pass


def test_icm_option_setter():
    """
        >>> rule = Rule()
        >>> rule.icm_option = 'a'
        Traceback (most recent call last):
        ValueError: Unknown value for icm_option: a
    """
    pass


def test_cm_option_setter():
    """
        >>> rule = Rule()
        >>> rule.cm_option = 'a'
        Traceback (most recent call last):
        ValueError: Unknown value for cm_option: a
    """
    pass


def test_is_not_iia():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 3, 2, 1],
        ...     [1, 0, 2, 3],
        ...     [2, 0, 1, 3],
        ...     [2, 1, 0, 3],
        ...     [3, 1, 2, 0],
        ... ])
        >>> rule = RulePlurality()(profile)
        >>> rule.is_not_iia_
        nan
    """
    pass


def test_compute_iia():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 1, 0],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleVeto()(profile)
        >>> rule.is_iia_
        False

        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 1, 2],
        ...     [1, 0, 2],
        ...     [1, 2, 0],
        ...     [1, 2, 0],
        ... ])
        >>> rule = RulePlurality()(profile)
        >>> rule.is_iia_
        True

        >>> profile = Profile(preferences_ut=[
        ...     [ 1. , -1. ,  0. ],
        ...     [ 1. ,  1. ,  0.5],
        ...     [-1. , -1. ,  0. ],
        ...     [ 1. ,  0.5,  1. ],
        ...     [ 0. , -1. ,  0. ],
        ... ], preferences_rk=[
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 0, 1],
        ...     [2, 0, 1],
        ... ])
        >>> rule = RuleMaximin()(profile)
        >>> rule.is_iia_
        True

        >>> profile = Profile(preferences_ut=[
        ...     [ 0.5,  0.5,  0. ],
        ...     [-0.5,  0.5,  0.5],
        ...     [-1. ,  0.5,  0. ],
        ...     [-1. ,  1. ,  1. ],
        ...     [-1. , -0.5, -0.5],
        ... ], preferences_rk=[
        ...     [1, 0, 2],
        ...     [1, 2, 0],
        ...     [1, 2, 0],
        ...     [2, 1, 0],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleExhaustiveBallot()(profile)
        >>> rule.is_iia_
        True

        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 4, 2, 3],
        ...     [0, 3, 2, 4, 1],
        ...     [1, 0, 4, 3, 2],
        ...     [1, 0, 4, 3, 2],
        ...     [2, 0, 4, 3, 1],
        ...     [4, 3, 1, 2, 0],
        ... ])
        >>> rule = RuleBaldwin()(profile)
        >>> rule.is_iia_
        True

        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2, 4, 3],
        ...     [0, 1, 3, 2, 4],
        ...     [0, 3, 2, 4, 1],
        ...     [1, 4, 3, 2, 0],
        ...     [2, 0, 1, 4, 3],
        ...     [4, 2, 3, 1, 0],
        ... ])
        >>> rule = RuleIRVAverage()(profile)
        >>> rule.is_iia_
        True

        >>> profile = Profile(preferences_ut=[
        ...     [ 0. ,  0. , -0.5],
        ...     [ 0.5, -1. ,  0. ],
        ...     [-0.5, -1. , -1. ],
        ...     [ 0. ,  0.5, -1. ],
        ...     [-1. ,  0.5, -1. ],
        ...     [-1. , -0.5,  0.5],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [1, 2, 0],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RulePlurality()(profile)
        >>> rule.is_iia_
        True

        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5,  0. ],
        ...     [ 0.5,  1. , -1. ],
        ...     [-0.5,  0. , -1. ],
        ...     [ 0.5,  0.5,  1. ],
        ...     [ 0. , -0.5,  0.5],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleMaximin()(profile)
        >>> rule.is_iia_
        True
    """
    pass


def test_compute_iia_aux():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleMajorityJudgment()(profile)
        >>> rule.is_iia_
        True

        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 1, 0],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleMajorityJudgment()(profile)
        >>> rule.is_iia_
        False
    """
    pass


def test_voters_im():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RulePlurality()(profile)
        >>> rule.voters_im_
        array([0., 0., 0., 0., 0.])
    """
    pass


def test_im_main_work_v_exact_rankings():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [1, 2, 0],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleNanson(im_option='exact')(profile)
        >>> rule.is_im_
        True

        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 2, 0],
        ...     [1, 2, 0],
        ...     [2, 0, 1],
        ... ])
        >>> rule = RuleNanson(im_option='exact')(profile)
        >>> rule.v_im_for_c_
        array([[0., 0., 0.],
               [0., 0., 0.],
               [0., 0., 1.],
               [0., 0., 1.],
               [0., 0., 0.]])

        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [1, 0, 2],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleNanson(im_option='exact')(profile)
        >>> rule.v_im_for_c_
        array([[1., 0., 0.],
               [0., 0., 0.],
               [0., 0., 0.],
               [0., 0., 0.],
               [0., 0., 0.]])
        """
    pass


def test_compute_im():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleNanson()(profile)
        >>> rule.voters_im_
        array([nan, nan,  0.,  0.,  0.])

        >>> profile = Profile(preferences_ut=[
        ...     [ 1. ,  1. ,  0. , -0.5],
        ...     [ 0.5,  1. , -0.5,  0. ],
        ...     [ 0. , -1. ,  1. , -1. ],
        ...     [-0.5,  0. ,  0. , -0.5],
        ... ], preferences_rk=[
        ...     [0, 1, 2, 3],
        ...     [1, 0, 3, 2],
        ...     [2, 0, 3, 1],
        ...     [2, 1, 3, 0],
        ... ])
        >>> rule = RuleIRVAverage(im_option='exact')(profile)
        >>> rule.voters_im_
        array([1., 1., 0., 0.])

        >>> profile = Profile(preferences_ut=[
        ...     [ 0.5, -0.5, -0.5, -1. ],
        ...     [ 1. ,  0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. , -1. ],
        ...     [-0.5,  0. ,  0.5, -0.5],
        ...     [ 0.5,  0.5,  0.5,  1. ],
        ...     [-1. , -1. ,  0. ,  0.5],
        ... ], preferences_rk=[
        ...     [0, 1, 2, 3],
        ...     [0, 2, 1, 3],
        ...     [2, 0, 1, 3],
        ...     [2, 1, 3, 0],
        ...     [3, 1, 0, 2],
        ...     [3, 2, 0, 1],
        ... ])
        >>> rule = RuleCondorcetSumDefeats(im_option='exact')(profile)
        >>> rule.candidates_im_
        array([1., 0., 0., 0.])
    """
    pass


def test_tm_preliminary_checks_general():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [1, 2, 0],
        ...     [2, 1, 0],
        ...     [2, 1, 0],
        ...     [2, 1, 0],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleRankedPairs()(profile)
        >>> rule.candidates_tm_
        array([0., 0., 0.])

        >>> profile = Profile(preferences_rk=[
        ...     [1, 0, 2],
        ...     [1, 0, 2],
        ...     [1, 2, 0],
        ...     [1, 2, 0],
        ...     [1, 2, 0],
        ... ])
        >>> rule = RuleBorda()(profile)
        >>> rule.candidates_tm_
        array([0., 0., 0.])

        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5,  1. , -1. ],
        ...     [ 1. ,  0.5,  1. ,  0.5],
        ...     [ 0. , -0.5, -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [2, 0, 1, 3],
        ...     [2, 0, 1, 3],
        ...     [3, 0, 1, 2],
        ... ])
        >>> rule = RuleSchulze()(profile)
        >>> rule.candidates_tm_
        array([0., 0., 0., 0.])

        >>> profile = Profile(preferences_ut=[
        ...     [ 1. ,  0.5, -0.5,  0. ],
        ...     [ 1. , -1. , -0.5, -0.5],
        ...     [ 0. , -1. , -1. ,  0. ],
        ...     [ 0. ,  0.5,  0.5,  0. ],
        ...     [-1. , -1. ,  0.5,  0. ],
        ...     [-0.5,  0. ,  1. ,  0.5],
        ... ], preferences_rk=[
        ...     [0, 1, 3, 2],
        ...     [0, 2, 3, 1],
        ...     [0, 3, 2, 1],
        ...     [1, 2, 0, 3],
        ...     [2, 3, 0, 1],
        ...     [2, 3, 1, 0],
        ... ])
        >>> rule = RuleMaximin()(profile)
        >>> rule.candidates_tm_
        array([0., 0., 0., 0.])

        >>> profile = Profile(preferences_ut=[
        ...     [-0.5, -1. ],
        ...     [ 0. ,  0.5],
        ... ], preferences_rk=[
        ...     [0, 1],
        ...     [1, 0],
        ... ])
        >>> rule = RuleIRVDuels()(profile)
        >>> rule.candidates_tm_
        array([0., 0.])
    """
    pass


def test_tm_main_work_c_lazy():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [1, 0, 2],
        ...     [1, 2, 0],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleNanson(tm_option='lazy')(profile)
        >>> rule.is_tm_
        nan
    """
    pass


def test_um_preliminary_checks_general():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [1, 2, 0],
        ...     [1, 2, 0],
        ... ])
        >>> rule = RuleNanson()(profile)
        >>> rule.candidates_um_
        array([0., 0., 0.])

        >>> profile = Profile(preferences_rk=[
        ...     [0, 3, 2, 1],
        ...     [1, 2, 0, 3],
        ...     [1, 2, 3, 0],
        ...     [2, 0, 1, 3],
        ...     [3, 2, 0, 1],
        ... ])
        >>> rule = RuleSchulze()(profile)
        >>> rule.candidates_um_
        array([ 0.,  1.,  0., nan])

        >>> profile = Profile(preferences_ut=[
        ...     [ 0.5, -0.5,  0.5],
        ...     [ 0. ,  0.5, -1. ],
        ...     [ 1. ,  1. ,  0.5],
        ...     [-1. ,  0. , -0.5],
        ...     [ 0. ,  0.5,  1. ],
        ... ], preferences_rk=[
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [1, 0, 2],
        ...     [1, 2, 0],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleMaximin()(profile)
        >>> rule.candidates_um_
        array([0., 0., 0.])

        >>> profile = Profile(preferences_ut=[
        ...     [ 0. ,  0. ,  0. ],
        ...     [ 0.5,  0. , -0.5],
        ...     [ 0. , -1. ,  0. ],
        ...     [ 0. ,  0.5,  0. ],
        ...     [-0.5,  1. , -0.5],
        ...     [-0.5, -0.5,  0. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [1, 0, 2],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleSchulze()(profile)
        >>> rule.candidates_um_
        array([0., 0., 0.])

        >>> profile = Profile(preferences_ut=[
        ...     [ 1. , -0.5, -1. ],
        ...     [ 1. , -1. , -0.5],
        ...     [ 1. ,  0. ,  0. ],
        ...     [-0.5,  0. , -0.5],
        ...     [-0.5,  0. , -0.5],
        ...     [-1. ,  0.5,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [1, 0, 2],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleExhaustiveBallot()(profile)
        >>> rule.candidates_um_
        array([0., 0., 0.])

        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 3, 2],
        ...     [0, 2, 1, 3],
        ...     [1, 0, 3, 2],
        ...     [2, 0, 3, 1],
        ...     [3, 0, 1, 2],
        ... ])
        >>> rule = RuleICRV()(profile)
        >>> rule.is_um_
        False
    """
    pass


def test_um_preliminary_checks_c():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 3, 2],
        ...     [0, 2, 3, 1],
        ...     [1, 0, 2, 3],
        ...     [1, 3, 2, 0],
        ...     [3, 1, 2, 0],
        ... ])
        >>> rule = RuleBaldwin()(profile)
        >>> rule.candidates_um_
        array([nan,  0.,  0., nan])

        >>> profile = Profile(preferences_ut=[
        ...     [ 1. ,  0.5,  0.5],
        ...     [-0.5,  0.5, -0.5],
        ...     [ 0.5,  1. ,  0.5],
        ...     [ 0. ,  0. , -1. ],
        ...     [ 0.5, -1. ,  1. ],
        ...     [ 0.5, -1. ,  0.5],
        ... ], preferences_rk=[
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [1, 0, 2],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 0, 1],
        ... ])
        >>> rule = RuleExhaustiveBallot()(profile)
        >>> rule.candidates_um_
        array([ 1.,  0., nan])
    """
    pass


def test_um_main_work_c_exact():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 2, 1],
        ...     [0, 2, 1],
        ...     [1, 2, 0],
        ...     [1, 2, 0],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleNanson(um_option='exact')(profile)
        >>> rule.is_um_
        True
    """
    pass


def test_icm_preliminary_checks_general():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 2, 1],
        ...     [1, 2, 0],
        ...     [1, 2, 0],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RulePlurality()(profile)
        >>> rule.is_icm_c_(1)
        False

        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1]
        ... ])
        >>> rule = RuleVeto()(profile)
        >>> rule.is_icm_
        False
    """
    pass


def test_icm_preliminary_checks_c():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 2, 1],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 1, 0],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RulePlurality()(profile)
        >>> rule.is_icm_c_(1)
        True

        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [1, 0, 2],
        ...     [1, 2, 0],
        ...     [2, 0, 1],
        ...     [2, 0, 1],
        ... ])
        >>> rule = RuleRangeVoting()(profile)
        >>> rule.is_icm_c_(1)
        False

        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2, 3],
        ...     [1, 0, 3, 2],
        ...     [1, 2, 0, 3],
        ...     [1, 3, 0, 2],
        ...     [2, 0, 1, 3],
        ...     [2, 0, 3, 1],
        ... ])
        >>> rule = RuleExhaustiveBallot()(profile)
        >>> rule.is_icm_
        True
    """
    pass


def test_candidates_icm():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RulePlurality()(profile)
        >>> rule.candidates_icm_
        array([0., 0., 1.])
    """
    pass


def test_icm_conclude_c():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 1, 0],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleBorda()(profile)
        >>> rule.is_icm_c_(1)
        True
    """
    pass


def test_compute_icm():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [1, 0, 2],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleBorda()(profile)
        >>> rule.is_icm_
        True
    """
    pass


def test_cm_preliminary_checks_general():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 1, 2],
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [0, 2, 1],
        ... ])
        >>> rule = RuleMajorityJudgment()(profile)
        >>> rule.is_cm_
        False

        >>> profile = Profile(preferences_rk=[
        ...     [0, 2, 3, 1],
        ...     [1, 2, 3, 0],
        ...     [2, 0, 3, 1],
        ...     [2, 1, 3, 0],
        ...     [3, 2, 1, 0],
        ... ])
        >>> rule = RuleBaldwin()(profile)
        >>> rule.is_cm_
        False

        >>> profile = Profile(preferences_ut=[
        ...     [ 0.5, -0.5,  0. ],
        ...     [-0.5,  0.5, -1. ],
        ...     [-0.5,  0. ,  0. ],
        ...     [-0.5,  1. ,  0. ],
        ...     [-1. , -0.5,  1. ],
        ... ], preferences_rk=[
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [1, 2, 0],
        ...     [1, 2, 0],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleKemeny()(profile)
        >>> rule.is_cm_
        False

        >>> profile = Profile(preferences_rk=[
        ...     [0, 2, 1],
        ...     [0, 2, 1],
        ...     [0, 2, 1],
        ...     [1, 2, 0],
        ...     [1, 2, 0],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleSchulze()(profile)
        >>> rule.is_cm_
        False

        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [1, 2, 0],
        ...     [1, 2, 0],
        ...     [2, 0, 1],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleIRVAverage()(profile)
        >>> rule.is_cm_
        True

        >>> profile = Profile(preferences_ut=[
        ...     [ 1. , -1. , -1. ],
        ...     [ 1. ,  0.5, -1. ],
        ...     [ 0.5, -0.5,  0.5],
        ...     [-0.5,  0. , -0.5],
        ...     [ 0.5,  1. ,  0.5],
        ...     [-0.5,  0.5,  0.5],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [1, 0, 2],
        ...     [1, 2, 0],
        ... ])
        >>> rule = RuleIRVDuels()(profile)
        >>> rule.is_cm_
        False
    """
    pass


def test_cm_preliminary_checks_c():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [1, 3, 0, 2],
        ...     [2, 1, 0, 3],
        ...     [2, 3, 1, 0],
        ...     [3, 0, 1, 2],
        ...     [3, 0, 1, 2],
        ... ])
        >>> rule = RuleNanson()(profile)
        >>> rule.is_cm_
        nan

        >>> profile = Profile(preferences_rk=[
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [1, 2, 0],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleBaldwin()(profile)
        >>> rule.is_cm_c_(1)
        True
    """
    pass


def test_cm_main_work_c_exact():
    """
        >>> profile = Profile(preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [1, 0, 2],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleNanson(cm_option='exact')(profile)
        >>> rule.is_cm_
        True
    """
    pass


# noinspection PyProtectedMember
def test_reached_uncovered_code():
    """
        >>> OPTIONS.ERROR_WHEN_UNCOVERED_CODE = True
        >>> profile = Profile(preferences_rk=[[0, 1, 2], [1, 0, 2]])
        >>> rule = RulePlurality()(profile)
        >>> rule._example_reached_uncovered_code()
        Traceback (most recent call last):
        AssertionError: Uncovered portion of code.
    """
    pass


def test_is_im_c_with_voters():
    """
        >>> profile = Profile(preferences_ut=[
        ...     [ 1. ,  0.5,  0. ],
        ...     [-1. ,  1. ,  0. ],
        ...     [-1. , -0.5,  0. ],
        ...     [-0.5, -0.5,  0. ],
        ...     [-0.5, -0.5,  1. ],
        ... ])
        >>> rule = RuleRangeVoting()(profile)
        >>> rule.is_im_v_with_candidates_(0)
        (False, array([0., 0., 0.]))
        >>> rule.is_im_c_with_voters_(0)
        (False, array([0., 0., 0., 0., 0.]))
    """
    pass


def test_temp():
    """

    """
    pass



# -*- coding: utf-8 -*-
"""
Created on 16 jul. 2021, 13:57
Copyright François Durand 2014-2021
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from svvamp.rules.rule import Rule
from svvamp.utils.util_cache import cached_property
from svvamp.preferences.profile import Profile


class RuleBlack(Rule):
    """
    Black rule

    Notes
    -----
    Black does not :attr:`meets_condorcet_c_ut_rel`:

        >>> profile = Profile(preferences_ut=[
        ...     [-1. , -0.5, -0.5],
        ...     [-1. , -0.5,  0. ],
        ... ], preferences_rk=[
        ...     [1, 2, 0],
        ...     [2, 1, 0],
        ... ])
        >>> RuleBlack()(profile).w_
        1
        >>> profile.condorcet_winner_ut_rel
        2

    Black does not :attr:`meets_majority_favorite_c_ut_ctb`:

        >>> profile = Profile(preferences_ut=[
        ...     [ 0.5, -1. ,  0. ],
        ...     [-1. ,  0. ,  0.5],
        ... ], preferences_rk=[
        ...     [0, 2, 1],
        ...     [2, 1, 0],
        ... ])
        >>> RuleBlack()(profile).w_
        2
        >>> profile.majority_favorite_ut_ctb
        0
    """

    full_name = 'Black'
    abbreviation = 'Bla'

    options_parameters = Rule.options_parameters.copy()
    options_parameters['icm_option'] = {'allowed': ['exact'], 'default': 'exact'}

    def __init__(self, **kwargs):
        super().__init__(
            with_two_candidates_reduces_to_plurality=True, is_based_on_rk=True,
            precheck_icm=False,
            log_identity="BLACK", **kwargs
        )

    # %% Count the ballots

    @cached_property
    def w_(self):
        self.mylog("Compute winner", 1)
        if self.profile_.exists_condorcet_winner_rk:
            return self.profile_.condorcet_winner_rk
        else:
            return int(np.argmax(self.profile_.borda_score_c_rk))

    @cached_property
    def scores_(self):
        self.mylog("Compute scores", 1)
        scores_condorcet = np.zeros(self.profile_.n_c)
        if self.profile_.exists_condorcet_winner_rk:
            scores_condorcet[self.profile_.condorcet_winner_rk] = 1
        scores_borda = self.profile_.borda_score_c_rk
        return np.array(scores_condorcet, scores_borda)

    @cached_property
    def candidates_by_scores_best_to_worst_(self):
        self.mylog("Compute candidates_by_scores_best_to_worst", 1)
        return sorted(
            range(self.profile_.n_c),
            key=lambda c: list(self.scores_[:, c]),
            reverse=True
        )

    # %% Manipulation criteria of the voting system

    @cached_property
    def meets_condorcet_c_rk(self):
        return True

    @cached_property
    def meets_ignmc_c_ctb(self):
        return True



from svvamp.utils.misc import initialize_random_seeds
from svvamp import GeneratorProfileVMFHypercircle


def test():
    """
    When there is one group only, you may specify `vmf_probability` but it is ignored:

        >>> initialize_random_seeds()
        >>> generator = GeneratorProfileVMFHypercircle(n_v=5, n_c=3, vmf_concentration=10, vmf_probability=1)
        >>> profile = generator()
        >>> profile.preferences_ut
        array([[ 0.67886167, -0.73231774,  0.05345607],
               [ 0.62834837, -0.76570906,  0.1373607 ],
               [ 0.71859954, -0.6950244 , -0.02357514],
               [ 0.49333272, -0.81010857,  0.31677584],
               [ 0.63042345, -0.76457205,  0.1341486 ]])
    """
    pass



# -*- coding: utf-8 -*-
"""
Created on 7 dec. 2018, 16:35
Copyright François Durand 2014-2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from svvamp.rules.rule import Rule
from svvamp.utils.util_cache import cached_property
from svvamp.preferences.profile import Profile


class RuleIteratedBucklin(Rule):
    """Iterated Bucklin method.

    Examples
    --------
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleIteratedBucklin()(profile)
        >>> rule.demo_results_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        ************************
        *                      *
        *   Election Results   *
        *                      *
        ************************
        <BLANKLINE>
        ***************
        *   Results   *
        ***************
        profile_.preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        profile_.preferences_rk (reminder) =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        ballots =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        scores =
        [[0.83333333 0.66666667 0.66666667]
         [1.66666667 1.                inf]]
        candidates_by_scores_best_to_worst
        [0, 1, 2]
        scores_best_to_worst
        [[0.83333333 0.66666667 0.66666667]
         [1.66666667 1.                inf]]
        w = 0
        score_w = [0.83333333 1.66666667]
        total_utility_w = 1.0
        <BLANKLINE>
        *********************************
        *   Condorcet efficiency (rk)   *
        *********************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_rk_ctb = 0
        w_is_condorcet_winner_rk_ctb = True
        w_is_not_condorcet_winner_rk_ctb = False
        w_missed_condorcet_winner_rk_ctb = False
        <BLANKLINE>
        condorcet_winner_rk = 0
        w_is_condorcet_winner_rk = True
        w_is_not_condorcet_winner_rk = False
        w_missed_condorcet_winner_rk = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (relative)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_ut_rel_ctb = 0
        w_is_condorcet_winner_ut_rel_ctb = True
        w_is_not_condorcet_winner_ut_rel_ctb = False
        w_missed_condorcet_winner_ut_rel_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_rel = 0
        w_is_condorcet_winner_ut_rel = True
        w_is_not_condorcet_winner_ut_rel = False
        w_missed_condorcet_winner_ut_rel = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (absolute)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_admissible_candidates =
        [ True False False]
        w_is_condorcet_admissible = True
        w_is_not_condorcet_admissible = False
        w_missed_condorcet_admissible = False
        <BLANKLINE>
        weak_condorcet_winners =
        [ True False False]
        w_is_weak_condorcet_winner = True
        w_is_not_weak_condorcet_winner = False
        w_missed_weak_condorcet_winner = False
        <BLANKLINE>
        condorcet_winner_ut_abs_ctb = 0
        w_is_condorcet_winner_ut_abs_ctb = True
        w_is_not_condorcet_winner_ut_abs_ctb = False
        w_missed_condorcet_winner_ut_abs_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_abs = 0
        w_is_condorcet_winner_ut_abs = True
        w_is_not_condorcet_winner_ut_abs = False
        w_missed_condorcet_winner_ut_abs = False
        <BLANKLINE>
        resistant_condorcet_winner = nan
        w_is_resistant_condorcet_winner = False
        w_is_not_resistant_condorcet_winner = True
        w_missed_resistant_condorcet_winner = False
        >>> rule.demo_manipulation_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        *****************************
        *                           *
        *   Election Manipulation   *
        *                           *
        *****************************
        <BLANKLINE>
        *********************************************
        *   Basic properties of the voting system   *
        *********************************************
        with_two_candidates_reduces_to_plurality =  True
        is_based_on_rk =  True
        is_based_on_ut_minus1_1 =  False
        meets_iia =  False
        <BLANKLINE>
        ****************************************************
        *   Manipulation properties of the voting system   *
        ****************************************************
        Condorcet_c_ut_rel_ctb (False)     ==>     Condorcet_c_ut_rel (False)
         ||                                                               ||
         ||     Condorcet_c_rk_ctb (False) ==> Condorcet_c_rk (False)     ||
         ||           ||               ||       ||             ||         ||
         V            V                ||       ||             V          V
        Condorcet_c_ut_abs_ctb (False)     ==>     Condorcet_ut_abs_c (False)
         ||                            ||       ||                        ||
         ||                            V        V                         ||
         ||       maj_fav_c_rk_ctb (False) ==> maj_fav_c_rk (True)        ||
         ||           ||                                       ||         ||
         V            V                                        V          V
        majority_favorite_c_ut_ctb (False) ==> majority_favorite_c_ut (True)
         ||                                                               ||
         V                                                                V
        IgnMC_c_ctb (False)                ==>                IgnMC_c (True)
         ||                                                               ||
         V                                                                V
        InfMC_c_ctb (False)                ==>                InfMC_c (True)
        <BLANKLINE>
        *****************************************************
        *   Independence of Irrelevant Alternatives (IIA)   *
        *****************************************************
        w (reminder) = 0
        is_iia = True
        log_iia: iia_subset_maximum_size = 2.0
        example_winner_iia = nan
        example_subset_iia = nan
        <BLANKLINE>
        **********************
        *   c-Manipulators   *
        **********************
        w (reminder) = 0
        preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        v_wants_to_help_c =
        [[False False False]
         [False False False]
         [False False False]
         [False False  True]
         [False False  True]]
        <BLANKLINE>
        ************************************
        *   Individual Manipulation (IM)   *
        ************************************
        is_im = nan
        log_im: im_option = lazy
        candidates_im =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************
        *   Trivial Manipulation (TM)   *
        *********************************
        is_tm = False
        log_tm: tm_option = exact
        candidates_tm =
        [0. 0. 0.]
        <BLANKLINE>
        ********************************
        *   Unison Manipulation (UM)   *
        ********************************
        is_um = nan
        log_um: um_option = lazy
        candidates_um =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************************
        *   Ignorant-Coalition Manipulation (ICM)   *
        *********************************************
        is_icm = False
        log_icm: icm_option = lazy
        candidates_icm =
        [0. 0. 0.]
        necessary_coalition_size_icm =
        [0. 5. 3.]
        sufficient_coalition_size_icm =
        [0. 6. 4.]
        <BLANKLINE>
        ***********************************
        *   Coalition Manipulation (CM)   *
        ***********************************
        is_cm = nan
        log_cm: cm_option = lazy, um_option = lazy, icm_option = lazy, tm_option = exact
        candidates_cm =
        [ 0.  0. nan]
        necessary_coalition_size_cm =
        [0. 0. 1.]
        sufficient_coalition_size_cm =
        [0. 6. 4.]

    Notes
    -----
    The candidate with least *adjusted median Borda score* (cf. below) is eliminated. Then the new Borda scores are
    computed. Etc. Ties are broken in favor of lower-index candidates: in case of a tie, the candidate with highest
    index is eliminated.

    Adjusted median Borda score:

        Let ``med_c`` be the median Borda score for candidate ``c``. Let ``x_c`` the number of voters who put a lower
        Borda score to ``c``. Then ``c``'s adjusted median is ``med_c - x_c / (n_v + 1)``.

        If ``med_c > med_d``, then it is also true for the adjusted median. If ``med_c = med_d``, then ``c`` has a
        better adjusted median iff ``x_c < x_d``, i.e. if more voters give to ``c`` the Borda score ``med_c`` or
        higher.

        So, the best candidate by adjusted median is the :class:`~Bucklin` winner. Here, at each round,
        we eliminate the candidate with lowest adjusted median Borda score, which justifies the name of "Iterated
        Bucklin method".

    Unlike Baldwin method (= Iterated Borda), Iterated Bucklin does not meet the Condorcet criterion. Indeed,
    a Condorcet winner may have the (strictly) worst median ranking.

    * :meth:`is_cm_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    * :meth:`is_icm_`: The algorithm from superclass :class:`Rule` is polynomial and has a window of error of 1
      manipulator.
    * :meth:`is_im_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    * :meth:`is_iia`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    * :meth:`is_tm_`: Exact in polynomial time.
    * :meth:`is_um_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    """

    full_name = 'Iterated Bucklin'
    abbreviation = 'IB'

    options_parameters = Rule.options_parameters.copy()

    def __init__(self, **kwargs):
        super().__init__(
            with_two_candidates_reduces_to_plurality=True, is_based_on_rk=True,
            log_identity="ITERATED_BUCKLIN", **kwargs
        )

    @cached_property
    def _count_ballots_(self):
        self.mylog("Count ballots", 1)
        scores = np.zeros((self.profile_.n_c - 1, self.profile_.n_c))
        worst_to_best = []
        is_eliminated = np.zeros(self.profile_.n_c, dtype=np.bool)
        preferences_borda_temp = np.copy(self.profile_.preferences_borda_rk)
        for r in range(self.profile_.n_c - 1):
            for c in range(self.profile_.n_c):
                if is_eliminated[c]:
                    scores[r, c] = np.inf
                    continue
                median = np.median(preferences_borda_temp[:, c])
                x = np.sum(np.less(preferences_borda_temp[:, c], median))
                scores[r, c] = median - x / (self.profile_.n_v + 1)
            loser = np.where(scores[r, :] == np.min(scores[r, :]))[0][-1]  # Tie-breaking: last index
            is_eliminated[loser] = True
            worst_to_best.append(loser)
            preferences_borda_temp[
                np.less(preferences_borda_temp, preferences_borda_temp[:, loser][:, np.newaxis])
            ] += 1
        w = np.argmin(is_eliminated)
        worst_to_best.append(w)
        candidates_by_scores_best_to_worst = worst_to_best[::-1]
        return {'scores': scores, 'w': w, 'candidates_by_scores_best_to_worst': candidates_by_scores_best_to_worst}

    @cached_property
    def scores_(self):
        """2d array of integers. ``scores[r, c]`` is the adjusted median Borda score of candidate ``c`` at
        elimination round ``r``.
        """
        return self._count_ballots_['scores']

    @cached_property
    def w_(self):
        return self._count_ballots_['w']

    @cached_property
    def candidates_by_scores_best_to_worst_(self):
        """1d array of integers. Candidates are sorted according to their order of elimination.
        """
        return self._count_ballots_['candidates_by_scores_best_to_worst']

    # noinspection PyProtectedMember
    @cached_property
    def _one_v_might_be_pivotal_(self):
        """
            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleIteratedBucklin()(profile)
            >>> rule._one_v_might_be_pivotal_
            False
        """
        scores_r = np.zeros(self.profile_.n_c)
        is_eliminated = np.zeros(self.profile_.n_c, dtype=np.bool)
        preferences_borda_temp = np.copy(self.profile_.preferences_borda_rk)
        for r in range(self.profile_.n_c - 1):
            self.mylogv("Pre-testing pivotality: round r =", r, 3)
            loser_sincere = (self.candidates_by_scores_best_to_worst_[-1 - r])
            self.mylogv("Pre-testing pivotality: loser_sincere =", loser_sincere, 3)
            # At worst, one manipulator can put one ``c`` from top to bottom or from bottom to top. Can this change the
            # loser of this round?
            for c in range(self.profile_.n_c):
                if is_eliminated[c]:
                    scores_r[c] = np.inf
                    continue
                if c == loser_sincere:
                    median = np.median(
                        np.concatenate((preferences_borda_temp[:, c], [self.profile_.n_c, self.profile_.n_c])))
                    x = sum(preferences_borda_temp[:, c] < median) - 1
                else:
                    median = np.median(np.concatenate((preferences_borda_temp[:, c], [1, 1])))
                    x = sum(preferences_borda_temp[:, c] < median) + 1
                scores_r[c] = median - x / (self.profile_.n_v + 1)
            loser = np.where(scores_r == scores_r.min())[0][-1]
            self.mylogv("Pre-testing pivotality: loser =", loser, 3)
            if loser != loser_sincere:
                self.mylog("There might be a pivotal voter", 3)
                return True
            is_eliminated[loser] = True
            preferences_borda_temp[
                np.less(preferences_borda_temp, preferences_borda_temp[:, loser][:, np.newaxis])
            ] += 1
        else:
            self.mylog("There is no pivotal voter", 3)
            return False

    @cached_property
    def v_might_im_for_c_(self):
        return np.full((self.profile_.n_v, self.profile_.n_c), self._one_v_might_be_pivotal_)

    @cached_property
    def meets_majority_favorite_c_rk(self):
        return True



# -*- coding: utf-8 -*-
"""
Created on 4 dec. 2018, 13:48
Copyright François Durand 2014-2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from svvamp.rules.rule import Rule
from svvamp.utils.util_cache import cached_property
from svvamp.preferences.profile import Profile


class RuleBorda(Rule):
    """Borda rule.

    Examples
    --------
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleBorda()(profile)
        >>> rule.demo_results_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        ************************
        *                      *
        *   Election Results   *
        *                      *
        ************************
        <BLANKLINE>
        ***************
        *   Results   *
        ***************
        profile_.preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        profile_.preferences_rk (reminder) =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        ballots =
        [[2 1 0]
         [2 0 1]
         [1 2 0]
         [1 0 2]
         [0 1 2]]
        scores =
        [6 4 5]
        candidates_by_scores_best_to_worst
        [0 2 1]
        scores_best_to_worst
        [6 5 4]
        w = 0
        score_w = 6
        total_utility_w = 1.0
        <BLANKLINE>
        *********************************
        *   Condorcet efficiency (rk)   *
        *********************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_rk_ctb = 0
        w_is_condorcet_winner_rk_ctb = True
        w_is_not_condorcet_winner_rk_ctb = False
        w_missed_condorcet_winner_rk_ctb = False
        <BLANKLINE>
        condorcet_winner_rk = 0
        w_is_condorcet_winner_rk = True
        w_is_not_condorcet_winner_rk = False
        w_missed_condorcet_winner_rk = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (relative)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_ut_rel_ctb = 0
        w_is_condorcet_winner_ut_rel_ctb = True
        w_is_not_condorcet_winner_ut_rel_ctb = False
        w_missed_condorcet_winner_ut_rel_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_rel = 0
        w_is_condorcet_winner_ut_rel = True
        w_is_not_condorcet_winner_ut_rel = False
        w_missed_condorcet_winner_ut_rel = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (absolute)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_admissible_candidates =
        [ True False False]
        w_is_condorcet_admissible = True
        w_is_not_condorcet_admissible = False
        w_missed_condorcet_admissible = False
        <BLANKLINE>
        weak_condorcet_winners =
        [ True False False]
        w_is_weak_condorcet_winner = True
        w_is_not_weak_condorcet_winner = False
        w_missed_weak_condorcet_winner = False
        <BLANKLINE>
        condorcet_winner_ut_abs_ctb = 0
        w_is_condorcet_winner_ut_abs_ctb = True
        w_is_not_condorcet_winner_ut_abs_ctb = False
        w_missed_condorcet_winner_ut_abs_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_abs = 0
        w_is_condorcet_winner_ut_abs = True
        w_is_not_condorcet_winner_ut_abs = False
        w_missed_condorcet_winner_ut_abs = False
        <BLANKLINE>
        resistant_condorcet_winner = nan
        w_is_resistant_condorcet_winner = False
        w_is_not_resistant_condorcet_winner = True
        w_missed_resistant_condorcet_winner = False
        >>> rule.demo_manipulation_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        *****************************
        *                           *
        *   Election Manipulation   *
        *                           *
        *****************************
        <BLANKLINE>
        *********************************************
        *   Basic properties of the voting system   *
        *********************************************
        with_two_candidates_reduces_to_plurality =  True
        is_based_on_rk =  True
        is_based_on_ut_minus1_1 =  False
        meets_iia =  False
        <BLANKLINE>
        ****************************************************
        *   Manipulation properties of the voting system   *
        ****************************************************
        Condorcet_c_ut_rel_ctb (False)     ==>     Condorcet_c_ut_rel (False)
         ||                                                               ||
         ||     Condorcet_c_rk_ctb (False) ==> Condorcet_c_rk (False)     ||
         ||           ||               ||       ||             ||         ||
         V            V                ||       ||             V          V
        Condorcet_c_ut_abs_ctb (False)     ==>     Condorcet_ut_abs_c (False)
         ||                            ||       ||                        ||
         ||                            V        V                         ||
         ||       maj_fav_c_rk_ctb (False) ==> maj_fav_c_rk (False)       ||
         ||           ||                                       ||         ||
         V            V                                        V          V
        majority_favorite_c_ut_ctb (False) ==> majority_favorite_c_ut (False)
         ||                                                               ||
         V                                                                V
        IgnMC_c_ctb (False)                ==>                IgnMC_c (False)
         ||                                                               ||
         V                                                                V
        InfMC_c_ctb (True)                 ==>                InfMC_c (True)
        <BLANKLINE>
        *****************************************************
        *   Independence of Irrelevant Alternatives (IIA)   *
        *****************************************************
        w (reminder) = 0
        is_iia = True
        log_iia: iia_subset_maximum_size = 2.0
        example_winner_iia = nan
        example_subset_iia = nan
        <BLANKLINE>
        **********************
        *   c-Manipulators   *
        **********************
        w (reminder) = 0
        preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        v_wants_to_help_c =
        [[False False False]
         [False False False]
         [False False False]
         [False False  True]
         [False False  True]]
        <BLANKLINE>
        ************************************
        *   Individual Manipulation (IM)   *
        ************************************
        is_im = False
        log_im: im_option = exact
        candidates_im =
        [0. 0. 0.]
        <BLANKLINE>
        *********************************
        *   Trivial Manipulation (TM)   *
        *********************************
        is_tm = False
        log_tm: tm_option = exact
        candidates_tm =
        [0. 0. 0.]
        <BLANKLINE>
        ********************************
        *   Unison Manipulation (UM)   *
        ********************************
        is_um = False
        log_um: um_option = exact
        candidates_um =
        [0. 0. 0.]
        <BLANKLINE>
        *********************************************
        *   Ignorant-Coalition Manipulation (ICM)   *
        *********************************************
        is_icm = False
        log_icm: icm_option = fast
        candidates_icm =
        [0. 0. 0.]
        necessary_coalition_size_icm =
        [0. 7. 5.]
        sufficient_coalition_size_icm =
        [0. 7. 5.]
        <BLANKLINE>
        ***********************************
        *   Coalition Manipulation (CM)   *
        ***********************************
        is_cm = False
        log_cm: cm_option = fast, um_option = exact, tm_option = exact
        candidates_cm =
        [0. 0. 0.]
        necessary_coalition_size_cm =
        [0. 2. 3.]
        sufficient_coalition_size_cm =
        [0. 2. 3.]

    Notes
    -----
    Voter ``v`` gives (:attr:`n_c` - 1) points to her top-ranked candidate, (:attr:`n_c` - 2) to the second, ..., 0 to
    the last. Ties are broken by natural order on the candidates (lower index wins).

    * :meth:`is_cm_`: Deciding CM is NP-complete.

        * :attr:`cm_option` = ``'fast'``: Zuckerman et al. (2009). This approximation algorithm is polynomial and has
          a window of error of 1 manipulator.
        * :attr:`cm_option` = ``'exact'``: Non-polynomial algorithm from superclass :class:`Rule`.

    * :meth:`is_icm_`: Algorithm is polynomial and has a window of error of 1 manipulator.
    * :meth:`is_im_`: Exact in polynomial time.
    * :meth:`is_iia_`: Exact in polynomial time.
    * :meth:`is_tm_`: Exact in polynomial time.
    * :meth:`is_um_`: Exact in polynomial time.

    References
    ----------
    'Algorithms for the coalitional manipulation problem', M. Zuckerman, A. Procaccia and J. Rosenschein, 2009.

    'Unweighted Coalitional Manipulation Under the Borda Rule is NP-Hard', Nadja Betzler, Rolf Niedermeier and
    Gerhard Woeginger, 2011.

    'Complexity of and algorithms for the manipulation of Borda, Nanson's and Baldwin's voting rules',
    Jessica Davies, George Katsirelos, Nina Narodytska, Toby Walsh and Lirong Xia, 2014.
    """

    full_name = 'Borda'
    abbreviation = 'Bor'

    options_parameters = Rule.options_parameters.copy()
    options_parameters.update({
        'im_option': {'allowed': ['exact'], 'default': 'exact'},
        'tm_option': {'allowed': ['exact'], 'default': 'exact'},
        'um_option': {'allowed': ['exact'], 'default': 'exact'},
        'icm_option': {'allowed': ['fast'], 'default': 'fast'},
        'cm_option': {'allowed': ['fast', 'exact'], 'default': 'fast'},
    })

    def __init__(self, **kwargs):
        super().__init__(
            with_two_candidates_reduces_to_plurality=True, is_based_on_rk=True,
            precheck_icm=False, log_identity="BORDA", **kwargs
        )

    @cached_property
    def ballots_(self):
        """2d array of integers. ``ballots[v, c]`` is the Borda score attributed to candidate ``c`` by voter ``v``
        (between 0 and ``n_c - 1``).
        """
        self.mylog("Compute ballots", 1)
        return self.profile_.preferences_borda_rk

    @cached_property
    def scores_(self):
        """1d array of integers. ``scores[c]`` is the total Borda score for candidate ``c``.
        """
        self.mylog("Compute scores", 1)
        return self.profile_.borda_score_c_rk

    @cached_property
    def meets_infmc_c_ctb(self):
        return True

    # %% Independence of Irrelevant Alternatives (IIA)

    @cached_property
    def _compute_iia_(self):
        """
            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [1, 0, 2],
            ...     [1, 2, 0],
            ...     [2, 0, 1],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleBorda()(profile)
            >>> rule.is_iia_
            False
        """
        self.mylog("Compute IIA", 1)
        # If we remove candidate ``d``, candidate ``c`` loses ``matrix_duels_rk[c, d]`` Borda points and ``w`` loses
        # ``matrix_duels_rk[w, d]``. So, candidate ``c`` gains the difference, when compared to ``w``.
        impact_removal_d_on_c = np.add(- self.profile_.matrix_duels_rk,
                                       self.profile_.matrix_duels_rk[self.w_, :][np.newaxis, :])
        impact_removal_d_on_c[np.diag_indices(self.profile_.n_c)] = 0
        impact_removal_d_on_c[:, self.w_] = 0  # Forbidden to remove w
        # For ``c``, any candidate ``d`` s.t. ``impact_removal_d_on_c[c, d] > 0`` is a good choice to remove.
        # ``pseudo_scores[c]`` is ``c``'s score minus ``w``'s score after these removals. After all these removals,
        # it may be that another candidate ``e`` will have a better score than ``c``, but we don't care: the only
        # important thing is to make ``w`` lose.
        pseudo_scores = (self.scores_ - self.score_w_ + np.sum(np.maximum(impact_removal_d_on_c, 0), 1))
        self.mylogv("pseudo_scores =", pseudo_scores, 2)
        # If ``pseudo_scores[c] > 0`` for some ``c``, then IIA is broken.
        best_c = np.argmax(pseudo_scores)
        self.mylogv("best_c =", best_c, 2)
        if pseudo_scores[best_c] + (best_c < self.w_) > 0:
            # Since we chose the best ``c``, we know that she is the winner in her optimal subset.
            return {'is_iia': False, 'example_winner_iia': best_c,
                    'example_subset_iia': np.logical_not(impact_removal_d_on_c[best_c, :] > 0)}
        else:
            return {'is_iia': True, 'example_winner_iia': np.nan, 'example_subset_iia': np.nan}

    # %% Individual manipulation (IM)

    def _im_preliminary_checks_general_subclass_(self):
        # If we change the ballot of only one ``c``-manipulator ``v``, then ``score[c] - score[w]`` increases by at
        # most ``n_c - 2``. Indeed, c gave at least 1 point of difference to ``c`` vs ``w`` (she strictly prefers ``c``
        # to ``w``), and now she can give  at most ``n_c - 1`` points of difference (with ``c`` on top and ``w`` at
        # the bottom).
        self.mylogv('Preliminary checks: scores =', self.scores_)
        self.mylogv('Preliminary checks: scores_w =', self.score_w_)
        temp = self.scores_ + self.profile_.n_c - 2 + (np.array(range(self.profile_.n_c)) < self.w_)
        self.mylogm('Preliminary checks: scores + n_c - 2 + (c < w) =', temp)
        # self.mylogm('Preliminary checks: possible candidates =', temp > self.scores[self.w])
        self._v_im_for_c[:, temp <= self.scores_[self.w_]] = False

    def _im_main_work_v_(self, v, c_is_wanted, nb_wanted_undecided, stop_if_true):
        """
            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [2, 0, 1],
            ...     [2, 1, 0],
            ... ])
            >>> rule = RuleBorda()(profile)
            >>> rule.v_im_for_c_
            array([[0., 0., 0.],
                   [0., 0., 0.],
                   [0., 0., 0.],
                   [0., 0., 1.],
                   [0., 0., 0.]])

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 2, 0],
            ...     [1, 2, 0],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleBorda()(profile)
            >>> rule.is_im_c_(0)
            True

            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [0, 2, 1],
            ...     [2, 0, 1],
            ...     [2, 0, 1],
            ...     [2, 1, 0],
            ... ])
            >>> rule = RuleBorda()(profile)
            >>> rule.is_im_c_(0)
            True

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [2, 0, 1],
            ...     [2, 1, 0],
            ... ])
            >>> rule = RuleBorda()(profile)
            >>> rule.is_im_c_(2)
            True
        """
        scores_without_v = self.scores_ - self.ballots_[v, :]
        self.mylogv('scores_without_v =', scores_without_v)
        candidates_by_decreasing_score = np.argsort(- scores_without_v, kind='mergesort')
        # ``ballot_balance`` is the ballot that ``v`` must give to balance the scores of all candidates the best she
        # can.
        ballot_balance = np.argsort(candidates_by_decreasing_score)
        self.mylogv('ballot_balance =', ballot_balance, 3)
        for c in range(self.profile_.n_c):
            if not c_is_wanted[c]:
                continue
            if not np.isneginf(self._v_im_for_c[v, c]):
                continue
            self.mylogv('Candidate c =', c, 3)
            # Compared to ``ballot_balance``, the candidates that ``v`` was putting better than ``c`` lost 1 point.
            ballot = ballot_balance - np.greater(ballot_balance, ballot_balance[c])
            # Voter v gives the maximal score to c.
            ballot[c] = self.profile_.n_c - 1
            self.mylogv('ballot =', ballot, 3)
            w_test = np.argmax(scores_without_v + ballot)
            self.mylogv('w_test =', w_test, 3)
            if w_test == c:
                self._v_im_for_c[v, c] = True
                self._candidates_im[c] = True
                self._voters_im[v] = True
                self._is_im = True
                self.mylog("IM found", 3)
                if stop_if_true:
                    return
            else:
                self._v_im_for_c[v, c] = False
            nb_wanted_undecided -= 1
            if nb_wanted_undecided == 0:
                return

    # %% Trivial Manipulation (TM)

    # Defined in the superclass.

    # %% Unison manipulation (UM)

    def _um_main_work_c_exact_(self, c):
        scores_test = np.sum(self.ballots_[np.logical_not(self.v_wants_to_help_c_[:, c]), :], 0)
        candidates_by_decreasing_score = np.argsort(-scores_test, kind='mergesort')
        # Balancing ballot: put candidates in the order of their current scores (least point to the most dangerous).
        ballot = np.argsort(candidates_by_decreasing_score)
        # Now put c on top. And modify other Borda scores accordingly on the ballot.
        ballot -= np.greater(ballot, ballot[c])
        ballot[c] = self.profile_.n_c - 1
        # New scores = old scores + n_manipulators * ballot.
        scores_test += np.multiply(self.profile_.matrix_duels_ut[c, self.w_], ballot)
        w_test = np.argmax(scores_test)
        self._candidates_um[c] = (w_test == c)

    # %% Ignorant-Coalition Manipulation (ICM)

    # noinspection PyUnusedLocal
    def _icm_main_work_c_fast_(self, c, optimize_bounds):
        # n_s: number of sincere voters (= counter-manipulators)
        n_s = self.profile_.n_v - self.profile_.matrix_duels_ut[c, self.w_]
        n_c = self.profile_.n_c
        # Necessary condition:
        # ^^^^^^^^^^^^^^^^^^^^
        # Denote ``n_m`` the number of manipulators for ``c``. Each manipulator gives ``n_c - 1`` points to ``c`` and an
        # average of ``(n_c - 2) / 2`` points to each other candidate. Let us note ``d`` the other candidate with
        # maximum ``score_m`` (from the manipulators).
        # ``score_m[c] = n_m * (n_c - 1)``.
        # ``score_m[d] >= n_m * (n_c - 2) / 2``.
        # At worst, each counter-manipulator will give ``n_c - 1`` points to ``d`` and 0 point to ``c``.
        # ``score_total[c] = n_m * (n_c - 1)``
        # ``score_total[d] >= n_m * (n_c - 2) / 2 + n_s * (n_c - 1)``
        # Since these scores may be half-integers, we will multiply them by 2 to apply the tie=breaking rule
        # correctly. We need:
        # ``2 * score_total[c] >= 2 * score_total[d] + (c > 0)``, which leads to:
        # ``n_m >= (2 * n_s * (n_c - 1) + (c > 0)) / n_c``.
        necessary = np.ceil((2 * n_s * (n_c - 1) + (c > 0)) / n_c)
        # Sufficient condition for n_m = 2 * p
        # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
        # Manipulators can give ``n_c - 1`` points to ``c``. Other candidates have balanced scores (ballot [n_c - 2,
        # ..., 0] paired with [0, ..., n_c - 2]).
        # ``score_m[c] = 2p (n_c - 1)``
        # ``score_m[d] = 2p (n_c - 2) / 2``
        # After counter-manipulation:
        # ``score_total[c] = 2p (n_c - 1)``
        # ``score_total[d] = 2p (n_c - 2) / 2 + n_s (n_c - 1)``
        # For ICM, it is sufficient that ``score_total[c] >= score_total[d] + (c > 0)``, which translates as before to:
        # ``p >= (n_s (n_c - 1) + (c > 0)) / n_c``.
        # N.B.: From the general inequality ``2 ceil(x) <= ceil (2x)``, it can be deduced that
        # ``sufficient_even - necessary <= 1``.
        sufficient_even = 2 * np.ceil((n_s * (n_c - 1) + (c > 0)) / n_c)
        # Sufficient condition for n_m = 2 * p + 1
        # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
        # Manipulators can give ``n_c - 1`` points to ``c``. With the first ``2p`` manipulators, other candidates
        # have balanced scores as before. In addition, the last manipulator gives ``n_c - 2`` points to the opponent
        # candidate with the higher index (so that she benefits less from t.b. rule).
        # ``score_m[c] = (2p + 1) (n_c - 1)``
        # ``score_m[d] = 2p (n_c - 2) / 2 + (n_c - 2) = (p + 1) (n_c - 2)``
        # After counter-manipulation:
        # ``score_total[c] = (2p + 1) (n_c - 1)``
        # ``score_total[d] = (p + 1) (n_c - 2) + n_s (n_c - 1)``
        # For ICM, it is sufficient that ``score_total[c] >= score_total[d] + (c == n_c - 1)`` (the tie-breaking rule
        # is against ``c`` only if she is the last candidate). This translates to:
        # ``p >= (n_s (n_c - 1) - 1 + (c == n_c - 1)) / n_c``.
        sufficient_odd = 1 + 2 * np.ceil((n_s * (n_c - 1) - 1 + (c == n_c-1)) / n_c)
        self._update_necessary(self._necessary_coalition_size_icm, c, necessary)
        self._update_sufficient(self._sufficient_coalition_size_icm, c, sufficient_even)
        self._update_sufficient(self._sufficient_coalition_size_icm, c, sufficient_odd)

    # %% Coalition Manipulation (CM)

    def _cm_main_work_c_fast_(self, c, optimize_bounds):
        """Do the main work in CM loop for candidate ``c``.

        * Try to improve bounds ``_sufficient_coalition_size_cm[c]`` and ``_necessary_coalition_size_cm[c]``.

        Algorithm: Zuckerman et al. (2009), Algorithms for the coalitional manipulation problem.

            >>> profile = Profile(preferences_rk=[
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [1, 0, 2],
            ...     [1, 0, 2],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleBorda()(profile)
            >>> rule.candidates_cm_
            array([0., 0., 0.])
        """
        scores_test = np.sum(self.ballots_[np.logical_not(self.v_wants_to_help_c_[:, c]), :], 0)
        # We add a tie-breaking term [(C-1)/C, (C-2)/C, ..., 0] to ease the computations.
        scores_test = scores_test + (np.array(range(self.profile_.n_c - 1, -1, -1)) / self.profile_.n_c)
        self.mylogv('CM: Further check: scores_test =', scores_test, 3)

        # A family of necessary conditions
        # For ``k`` in ``{1, ..., n_c - 1}``, let us note ``mean_k`` the mean of the ``k`` highest scores among other
        # candidates than ``c``. At best, we will have:
        # ``scores_tot[c] = scores_test[c] + n_m (n_c - 1)``
        # ``mean_k_tot = mean_k_test + n_m (k - 1) / 2``
        # We must have ``scores_tot[c] >= mean_k_tot`` so:
        # ``n_m >= 2 * (mean_k_test - scores_test[c]) / (2 n_c - k - 1)``
        increasing_scores = np.sort(scores_test[np.array(range(self.profile_.n_c)) != c])
        cum_sum = 0
        for k in range(1, self.profile_.n_c):
            cum_sum += increasing_scores[-k]
            necessary = np.ceil(2 * (cum_sum / k - scores_test[c]) / (2 * self.profile_.n_c - k - 1))
            self.mylogv('CM: Further check: k =', k)
            self.mylogv('CM: Further check: cum_sum =', cum_sum)
            self.mylogv('CM: Further check: necessary =', necessary)
            self._update_necessary(self._necessary_coalition_size_cm, c, necessary,
                                   'CM: Further check: necessary_coalition_size_cm =')
        # An opportunity to escape before real work
        if self._necessary_coalition_size_cm[c] == self._sufficient_coalition_size_cm[c]:
            return False
        if not optimize_bounds and self.profile_.matrix_duels_ut[c, self.w_] < self._necessary_coalition_size_cm[c]:
            # This is a quick escape: we have not optimized the bounds the best we could.
            return True

        # Now, the real work
        n_m = 0
        self.mylogv('CM: Fast algorithm: scores_test =', scores_test, 3)
        while True:
            n_m += 1
            # Balancing ballot: put candidates in the order of their current scores (least point to the most
            # dangerous).
            candidates_by_decreasing_score = np.argsort(- scores_test, kind='mergesort')
            ballot = np.argsort(candidates_by_decreasing_score)
            # Now put ``c`` on top. And modify other Borda scores accordingly on the ballot.
            ballot -= np.greater(ballot, ballot[c])
            ballot[c] = self.profile_.n_c - 1
            self.mylogv('CM: Fast algorithm: ballot =', ballot, 3)
            # New scores = old scores + ballot.
            scores_test += ballot
            self.mylogv('CM: Fast algorithm: scores_test =', scores_test, 3)
            w_test = np.argmax(scores_test)
            if w_test == c:
                sufficient = n_m
                self.mylogv('CM: Fast algorithm: sufficient =', sufficient)
                self._update_sufficient(self._sufficient_coalition_size_cm, c, sufficient,
                                        'CM: Fast algorithm: sufficient_coalition_size_cm =')
                # From Zuckerman et al.: the algorithm is optimal up to 1 voter. But since we do not have the same
                # tie-breaking rule, it makes 2 voters.
                self._update_necessary(self._necessary_coalition_size_cm, c, sufficient - 2,
                                       'CM: Fast algorithm: necessary_coalition_size_cm =')
                break

    def _cm_main_work_c_(self, c, optimize_bounds):
        """
            >>> profile = Profile(preferences_rk=[
            ...     [0, 1, 2],
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [0, 2, 1],
            ...     [2, 0, 1],
            ... ])
            >>> rule = RuleBorda(cm_option='exact')(profile)
            >>> rule.candidates_cm_
            array([0., 0., 0.])
        """
        is_quick_escape_fast = self._cm_main_work_c_fast_(c, optimize_bounds)
        if not self.cm_option == "exact":
            # With 'fast' option, we stop here anyway.
            return is_quick_escape_fast
        # From this point, we have necessarily the 'exact' option (which is, in fact, only an exhaustive exploration
        # with = ``n_m`` manipulators).
        is_quick_escape_exact = self._cm_main_work_c_exact_(c, optimize_bounds)
        return is_quick_escape_fast or is_quick_escape_exact



# -*- coding: utf-8 -*-
"""
Created on nov. 28, 2018, 14:47
Copyright François Durand 2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from svvamp.preferences.generator_profile import GeneratorProfile
from svvamp.preferences.profile import Profile


class GeneratorProfileLadder(GeneratorProfile):
    """Profile generator using the 'Ladder' model.

    Parameters
    ----------
    n_v : int
        Number of voters.
    n_c : int
        Number of candidates.
    n_rungs : int
        Number of rungs of the ladder.
    sort_voters : bool
        This argument is passed to :class:`Profile`.

    Notes
    -----
    Each utility ``preferences_ut[v, c]`` is drawn independently and uniformly between ``n_rungs`` values that
    divide the interval [-1, 1] equally. For example, if ``n_rungs = 21``, then values in {-1, -0.9, ..., 1} are used.

    The ordinal part of this distribution is **not** the Impartial Culture. Indeed, weak orders of preference come
    with non-zero probability. This model is interesting the study the effect of voters' ties.

    Examples
    --------
        >>> generator = GeneratorProfileLadder(n_v=10, n_c=3, n_rungs=5)
        >>> profile = generator()
        >>> profile.preferences_rk.shape
        (10, 3)
    """

    def __init__(self, n_v, n_c, n_rungs, sort_voters=True):
        self.n_v = n_v
        self.n_c = n_c
        self.n_rungs = n_rungs
        self.log_creation = ['Ladder', n_c, n_v, 'Number of rungs', n_rungs]
        super().__init__(sort_voters=sort_voters)

    def __call__(self):
        return Profile(
            preferences_ut=np.random.randint(self.n_rungs, size=(self.n_v, self.n_c)) * 2 / (self.n_rungs - 1) - 1,
            log_creation=self.log_creation)



def printm(message="Variable =", variable=None):
    """Print a message and the value of a variable, typically a matrix.

    This method is well suited for a matrix because it skips to next line before printing the variable.

    Parameters
    ----------
    message : str
    variable : object
        The variable to be displayed.

    Examples
    --------
        >>> import numpy as np
        >>> printm("A nice matrix:", np.array([[2, 7, 6], [9, 5, 1], [4, 3, 8]]))
        A nice matrix:
        [[2 7 6]
         [9 5 1]
         [4 3 8]]
    """
    print(message)
    print(variable)


def print_title(s):
    """Print a title nicely.

    Parameters
    ----------
    s : str

    Examples
    --------
        >>> print_title('A title')
        <BLANKLINE>
        ***************
        *   A title   *
        ***************
    """
    length = len(s)
    print('')
    print('*' * (length + 8))
    print('*   ' + s + '   *')
    print('*' * (length + 8))


def print_big_title(s):
    """Print a big title nicely.

    Examples
    --------
    s : str

    Examples
    --------
        >>> print_big_title('A big title')
        <BLANKLINE>
        *******************
        *                 *
        *   A big title   *
        *                 *
        *******************
    """
    length = len(s)
    print('')
    print('*' * (length + 8))
    print('*' + ' ' * (length + 6) + '*')
    print('*   ' + s + '   *')
    print('*' + ' ' * (length + 6) + '*')
    print('*' * (length + 8))



# -*- coding: utf-8 -*-
"""
Created on oct. 21, 2014, 09:54
Copyright François Durand 2014-2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from svvamp.utils.misc import initialize_random_seeds
from svvamp.utils.util_cache import cached_property
from svvamp.preferences.profile import preferences_ut_to_preferences_rk
from svvamp.preferences.profile import Profile


class ProfileSubsetCandidates(Profile):
    """Sub-profile for a subset of the original candidates

    Parameters
    ----------
    parent_profile : Profile
        The initial profile.
    candidates_subset : list
        Normally a list of candidates indexes, like [0, 2, 3], but a list of booleans like
        [True False True True False] should work as well. Candidates belonging to the subset.

    Notes
    -----
    N.B.: if candidates_subset is a list of integers, it must be sorted in ascending order.

    N.B.: in this object, candidates are re-numbered, for example [0, 1, 2]. So, if the winner of an election is w in
    the sub-profile, it corresponds to candidates_subset[w] in the parent profile (supposing that candidates_subset is
    given as a list of indexes, not as a list of booleans).

    Examples
    --------
        >>> initialize_random_seeds()
        >>> preferences_ut_test = np.random.randint(-5, 5, (10, 5))
        >>> parent_profile = Profile(preferences_ut=preferences_ut_test)
        >>> profile = ProfileSubsetCandidates(parent_profile=parent_profile, candidates_subset=[0, 1, 2])
        >>> profile.matrix_duels_ut
        array([[0, 5, 3],
               [3, 0, 2],
               [6, 7, 0]])
        >>> profile.matrix_duels_rk
        array([[0, 6, 3],
               [4, 0, 2],
               [7, 8, 0]])
        >>> profile.matrix_victories_ut_abs
        array([[0. , 0.5, 0. ],
               [0. , 0. , 0. ],
               [1. , 1. , 0. ]])
        >>> profile.matrix_victories_ut_abs_ctb
        array([[0., 1., 0.],
               [0., 0., 0.],
               [1., 1., 0.]])
        >>> profile.matrix_victories_ut_rel
        array([[0., 1., 0.],
               [0., 0., 0.],
               [1., 1., 0.]])
        >>> profile.matrix_victories_ut_rel_ctb
        array([[0., 1., 0.],
               [0., 0., 0.],
               [1., 1., 0.]])
        >>> profile.matrix_victories_rk
        array([[0., 1., 0.],
               [0., 0., 0.],
               [1., 1., 0.]])
        >>> profile.matrix_victories_rk_ctb
        array([[0., 1., 0.],
               [0., 0., 0.],
               [1., 1., 0.]])
        >>> profile.total_utility_c
        array([ 2, -8,  6])
    """

    def __init__(self, parent_profile, candidates_subset):
        self.parent_profile = parent_profile
        # noinspection PyTypeChecker
        self.candidates_subset = candidates_subset
        super().__init__(preferences_ut=parent_profile.preferences_ut[:, candidates_subset],
                         preferences_rk=preferences_ut_to_preferences_rk(
                             self.parent_profile.preferences_borda_rk[:, candidates_subset]),
                         sort_voters=False)

    # %% Matrix of duels

    @cached_property
    def matrix_duels_ut(self):
        return self.parent_profile.matrix_duels_ut[self.candidates_subset, :][:, self.candidates_subset]

    @cached_property
    def matrix_duels_rk(self):
        return self.parent_profile.matrix_duels_rk[self.candidates_subset, :][:, self.candidates_subset]

    @cached_property
    def matrix_victories_ut_abs(self):
        return self.parent_profile.matrix_victories_ut_abs[self.candidates_subset, :][:, self.candidates_subset]

    @cached_property
    def matrix_victories_ut_abs_ctb(self):
        return self.parent_profile.matrix_victories_ut_abs_ctb[self.candidates_subset, :][:, self.candidates_subset]

    @cached_property
    def matrix_victories_ut_rel(self):
        return self.parent_profile.matrix_victories_ut_rel[self.candidates_subset, :][:, self.candidates_subset]

    @cached_property
    def matrix_victories_ut_rel_ctb(self):
        return self.parent_profile.matrix_victories_ut_rel_ctb[self.candidates_subset, :][:, self.candidates_subset]

    @cached_property
    def matrix_victories_rk(self):
        return self.parent_profile.matrix_victories_rk[self.candidates_subset, :][:, self.candidates_subset]

    @cached_property
    def matrix_victories_rk_ctb(self):
        return self.parent_profile.matrix_victories_rk_ctb[self.candidates_subset, :][:, self.candidates_subset]

    # %% Total utilities

    @cached_property
    def total_utility_c(self):
        return self.parent_profile.total_utility_c[self.candidates_subset]



# -*- coding: utf-8 -*-
"""
Created on 4 dec. 2018, 11:42
Copyright François Durand 2014-2018
fradurand@gmail.com

This file is part of SVVAMP.

    SVVAMP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SVVAMP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SVVAMP.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np
from svvamp.rules.rule import Rule
from svvamp.utils.util_cache import cached_property
from svvamp.preferences.profile import Profile


class RuleBaldwin(Rule):
    """Baldwin method.

    Examples
    --------
        >>> profile = Profile(preferences_ut=[
        ...     [ 0. , -0.5, -1. ],
        ...     [ 1. , -1. ,  0.5],
        ...     [ 0.5,  0.5, -0.5],
        ...     [ 0.5,  0. ,  1. ],
        ...     [-1. , -1. ,  1. ],
        ... ], preferences_rk=[
        ...     [0, 1, 2],
        ...     [0, 2, 1],
        ...     [1, 0, 2],
        ...     [2, 0, 1],
        ...     [2, 1, 0],
        ... ])
        >>> rule = RuleBaldwin()(profile)
        >>> rule.demo_results_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        ************************
        *                      *
        *   Election Results   *
        *                      *
        ************************
        <BLANKLINE>
        ***************
        *   Results   *
        ***************
        profile_.preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        profile_.preferences_rk (reminder) =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        ballots =
        [[0 1 2]
         [0 2 1]
         [1 0 2]
         [2 0 1]
         [2 1 0]]
        scores =
        [[ 6.  4.  5.]
         [ 3. inf  2.]]
        candidates_by_scores_best_to_worst
        [0 2 1]
        scores_best_to_worst
        [[ 6.  5.  4.]
         [ 3.  2. inf]]
        w = 0
        score_w = [6. 3.]
        total_utility_w = 1.0
        <BLANKLINE>
        *********************************
        *   Condorcet efficiency (rk)   *
        *********************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_rk_ctb = 0
        w_is_condorcet_winner_rk_ctb = True
        w_is_not_condorcet_winner_rk_ctb = False
        w_missed_condorcet_winner_rk_ctb = False
        <BLANKLINE>
        condorcet_winner_rk = 0
        w_is_condorcet_winner_rk = True
        w_is_not_condorcet_winner_rk = False
        w_missed_condorcet_winner_rk = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (relative)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_winner_ut_rel_ctb = 0
        w_is_condorcet_winner_ut_rel_ctb = True
        w_is_not_condorcet_winner_ut_rel_ctb = False
        w_missed_condorcet_winner_ut_rel_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_rel = 0
        w_is_condorcet_winner_ut_rel = True
        w_is_not_condorcet_winner_ut_rel = False
        w_missed_condorcet_winner_ut_rel = False
        <BLANKLINE>
        ***************************************
        *   Condorcet efficiency (absolute)   *
        ***************************************
        w (reminder) = 0
        <BLANKLINE>
        condorcet_admissible_candidates =
        [ True False False]
        w_is_condorcet_admissible = True
        w_is_not_condorcet_admissible = False
        w_missed_condorcet_admissible = False
        <BLANKLINE>
        weak_condorcet_winners =
        [ True False False]
        w_is_weak_condorcet_winner = True
        w_is_not_weak_condorcet_winner = False
        w_missed_weak_condorcet_winner = False
        <BLANKLINE>
        condorcet_winner_ut_abs_ctb = 0
        w_is_condorcet_winner_ut_abs_ctb = True
        w_is_not_condorcet_winner_ut_abs_ctb = False
        w_missed_condorcet_winner_ut_abs_ctb = False
        <BLANKLINE>
        condorcet_winner_ut_abs = 0
        w_is_condorcet_winner_ut_abs = True
        w_is_not_condorcet_winner_ut_abs = False
        w_missed_condorcet_winner_ut_abs = False
        <BLANKLINE>
        resistant_condorcet_winner = nan
        w_is_resistant_condorcet_winner = False
        w_is_not_resistant_condorcet_winner = True
        w_missed_resistant_condorcet_winner = False
        >>> rule.demo_manipulation_(log_depth=0)  # doctest: +NORMALIZE_WHITESPACE
        <BLANKLINE>
        *****************************
        *                           *
        *   Election Manipulation   *
        *                           *
        *****************************
        <BLANKLINE>
        *********************************************
        *   Basic properties of the voting system   *
        *********************************************
        with_two_candidates_reduces_to_plurality =  True
        is_based_on_rk =  True
        is_based_on_ut_minus1_1 =  False
        meets_iia =  False
        <BLANKLINE>
        ****************************************************
        *   Manipulation properties of the voting system   *
        ****************************************************
        Condorcet_c_ut_rel_ctb (False)     ==>     Condorcet_c_ut_rel (False)
         ||                                                               ||
         ||     Condorcet_c_rk_ctb (True)  ==> Condorcet_c_rk (True)      ||
         ||           ||               ||       ||             ||         ||
         V            V                ||       ||             V          V
        Condorcet_c_ut_abs_ctb (True)      ==>     Condorcet_ut_abs_c (True)
         ||                            ||       ||                        ||
         ||                            V        V                         ||
         ||       maj_fav_c_rk_ctb (True)  ==> maj_fav_c_rk (True)        ||
         ||           ||                                       ||         ||
         V            V                                        V          V
        majority_favorite_c_ut_ctb (True)  ==> majority_favorite_c_ut (True)
         ||                                                               ||
         V                                                                V
        IgnMC_c_ctb (True)                 ==>                IgnMC_c (True)
         ||                                                               ||
         V                                                                V
        InfMC_c_ctb (True)                 ==>                InfMC_c (True)
        <BLANKLINE>
        *****************************************************
        *   Independence of Irrelevant Alternatives (IIA)   *
        *****************************************************
        w (reminder) = 0
        is_iia = True
        log_iia: iia_subset_maximum_size = 2.0
        example_winner_iia = nan
        example_subset_iia = nan
        <BLANKLINE>
        **********************
        *   c-Manipulators   *
        **********************
        w (reminder) = 0
        preferences_ut (reminder) =
        [[ 0.  -0.5 -1. ]
         [ 1.  -1.   0.5]
         [ 0.5  0.5 -0.5]
         [ 0.5  0.   1. ]
         [-1.  -1.   1. ]]
        v_wants_to_help_c =
        [[False False False]
         [False False False]
         [False False False]
         [False False  True]
         [False False  True]]
        <BLANKLINE>
        ************************************
        *   Individual Manipulation (IM)   *
        ************************************
        is_im = nan
        log_im: im_option = lazy
        candidates_im =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************
        *   Trivial Manipulation (TM)   *
        *********************************
        is_tm = False
        log_tm: tm_option = exact
        candidates_tm =
        [0. 0. 0.]
        <BLANKLINE>
        ********************************
        *   Unison Manipulation (UM)   *
        ********************************
        is_um = nan
        log_um: um_option = lazy
        candidates_um =
        [ 0.  0. nan]
        <BLANKLINE>
        *********************************************
        *   Ignorant-Coalition Manipulation (ICM)   *
        *********************************************
        is_icm = False
        log_icm: icm_option = exact
        candidates_icm =
        [0. 0. 0.]
        necessary_coalition_size_icm =
        [0. 6. 4.]
        sufficient_coalition_size_icm =
        [0. 6. 4.]
        <BLANKLINE>
        ***********************************
        *   Coalition Manipulation (CM)   *
        ***********************************
        is_cm = nan
        log_cm: cm_option = lazy, um_option = lazy, tm_option = exact
        candidates_cm =
        [ 0.  0. nan]
        necessary_coalition_size_cm =
        [0. 1. 2.]
        sufficient_coalition_size_cm =
        [0. 6. 4.]

    Notes
    -----
    Each voter provides a strict order of preference. The candidate with lowest Borda score is eliminated. Then the
    new Borda scores are computed. Etc. Ties are broken in favor of lower-index candidates: in case of a tie,
    the candidate with highest index is eliminated.

    Since a Condorcet winner has always a Borda score higher than average, Baldwin method meets the Condorcet
    criterion.

    * :meth:`is_cm_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    * :meth:`is_icm_`: Exact in polynomial time.
    * :meth:`is_im_`: Deciding IM is NP-complete. Non-polynomial or non-exact algorithms from superclass :class:`Rule`.
    * :meth:`is_iia_`: Exact in polynomial time.
    * :meth:`is_tm_`: Exact in polynomial time.
    * :meth:`is_um_`: Non-polynomial or non-exact algorithms from superclass :class:`Rule`.

    References
    ----------
    'Complexity of and algorithms for the manipulation of Borda, Nanson's and Baldwin's voting rules',
    Jessica Davies, George Katsirelos, Nina Narodytska, Toby Walsh and Lirong Xia, 2014.
    """

    full_name = 'Baldwin'
    abbreviation = 'Bal'

    options_parameters = Rule.options_parameters.copy()
    options_parameters['icm_option'] = {'allowed': ['exact'], 'default': 'exact'}

    def __init__(self, **kwargs):
        super().__init__(
            with_two_candidates_reduces_to_plurality=True, is_based_on_rk=True,
            precheck_icm=False, log_identity="BALDWIN", **kwargs
        )

    @cached_property
    def _count_ballots_(self):
        """
        Returns
        -------
        dict
            A dictionary with ``scores``, ``w``, ``candidates_by_scores_best_to_worst``
        """
        self.mylog("Count ballots", 1)
        scores = np.zeros((self.profile_.n_c - 1, self.profile_.n_c))
        worst_to_best = []
        preferences_borda_temp = self.profile_.preferences_borda_rk.astype(float)
        w = None
        for r in range(self.profile_.n_c - 1):  # Round r
            scores[r, :] = np.sum(preferences_borda_temp, 0)
            loser = np.where(scores[r, :] == np.min(scores[r, :]))[0][-1]  # Tie-breaking: higher index
            worst_to_best.append(loser)
            if r == self.profile_.n_c - 2:
                preferences_borda_temp[0, loser] = np.inf
                w = np.argmin(preferences_borda_temp[0, :])
                worst_to_best.append(w)
                break
            # Prepare for next round
            # If v was ranking c above loser, then c loses 1 point:
            preferences_borda_temp[preferences_borda_temp > preferences_borda_temp[:, loser][:, np.newaxis]] -= 1
            preferences_borda_temp[:, loser] = np.inf
        candidates_by_scores_best_to_worst = np.array(worst_to_best[::-1])
        return {'scores': scores, 'w': w, 'candidates_by_scores_best_to_worst': candidates_by_scores_best_to_worst}

    @cached_property
    def scores_(self):
        """2d array of integers. ``scores[r, c]`` is the Borda score of candidate ``c`` at elimination round ``r``.
        By convention, if candidate ``c`` does not participate to round ``r``, then ``scores[r, c] = numpy.inf``.
        """
        return self._count_ballots_['scores']

    @cached_property
    def w_(self):
        return self._count_ballots_['w']

    @cached_property
    def candidates_by_scores_best_to_worst_(self):
        """1d array of integers. ``candidates_by_scores_best_to_worst[-r]`` is the candidate eliminated at
        elimination round ``r``. By definition / convention, ``candidates_by_scores_best_to_worst[0]`` = :attr:`w_`.
        """
        return self._count_ballots_['candidates_by_scores_best_to_worst']

    @cached_property
    def meets_condorcet_c_rk_ctb(self):
        # Consider vtb case. If ``c`` is a Condorcet winner with vtb and ctb, then she has at worst ties in
        # ``matrix_duels_rk``, so she has at least the average Borda score. Can she be eliminated? For that, it is
        # necessary that all other candidates have exactly the average, i.e. ``matrix_duels_rk`` is a general tie.
        # Since she is Condorcet winner vtb/ctb, she must be candidate 0, so our tie-breaking rule eliminates another
        # candidate. Conclusion: this voting system meets Condorcet criterion vtb/ctb.
        #
        # However, let us consider the following example.
        # preferences_borda_ut:
        # [ 1   0 ]
        # [0.5 0.5]
        # [0.5 0.5]
        # ==> candidate 0 is a relative Condorcet winner.
        # preferences_borda_rk (with vtb):
        # [ 1   0 ]
        # [ 0   1 ]
        # [ 0   1 ]
        # ==> candidate 1 wins.
        # Conclusion: this voting system does not meet Condorcet criterion (ut/rel).
        return True
